import * as toml from "toml";
import Resource from "./resource";
import {
  UrlProvider,
  ModrinthProvider,
  CurseForgeProvider,
  GitHubProvider,
} from "./provider";
import { HashFormat, isValidHashFormat } from "./enums/hash-format";
import { isValidSide, type Side } from "./enums/side";

export interface Metafile {
  name: string;
  filename: string;
  side: Side;
  provider: UrlProvider;
  option: Option;
}

export interface Option {
  optional: boolean;
  default?: boolean;
  description?: string;
}

/**
 * Represents a file entry in the Packwiz index.
 */
export class IndexFileEntry {
  constructor(
    readonly file: Resource,
    readonly hash: string,
    readonly hashFormat: HashFormat,
    readonly alias?: string,
    readonly metafile: boolean = false,
    readonly preserve: boolean = false,
  ) {
    this.file = file instanceof Resource ? file : new Resource(file);
    this.hash = hash;
    this.hashFormat = isValidHashFormat(hashFormat);
    this.alias = alias;
    this.metafile = metafile;
    this.preserve = preserve;
  }

  async parse(): Promise<Metafile> {
    if (this.metafile === false) {
      throw new Error(
        "Cannot parse non-metafiles! Use `file.fetchContents()` instead to get the file contents.",
      );
    }
    const fileContent = await this.file.fetchContents();
    const parsed = toml.parse(fileContent);
    const side = isValidSide(parsed.side);

    const metafile: Metafile = {
      name: parsed.name,
      filename: parsed.filename,
      side: side,
      provider: this.parseProvider(parsed),
      option: {
        optional: parsed.option?.optional || false,
        default: parsed.option?.default,
        description: parsed.option?.description,
      },
    };

    return metafile;
  }

  private parseProvider(parsed: any): UrlProvider {
    if (parsed.update?.modrinth) {
      return new ModrinthProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        new Resource(parsed.download.url),
        parsed.update.modrinth["mod-id"],
        parsed.update.modrinth["version"],
      );
    } else if (parsed.update?.curseforge) {
      return new CurseForgeProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        parsed.download.mode,
        parsed.update.curseforge["file-id"],
        parsed.update.curseforge["project-id"],
      );
    } else if (parsed.update?.github) {
      return new GitHubProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        new Resource(parsed.download.url),
        parsed.update.github.branch,
        parsed.update.github.regex,
        parsed.update.github.slug,
        parsed.update.github.tag,
      );
    } else if (parsed.download) {
      return new UrlProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        new Resource(parsed.download.url),
      );
    } else {
      throw new Error("Unknown provider in TOML.");
    }
  }
}

/**
 * Represents the structure of a Packwiz index file, as well as providing its location.
 */
export interface PackwizIndex {
  location: Resource;
  hashFormat: HashFormat;
  files: IndexFileEntry[];
}

// TODO: parse pack.toml too instead of just the index file
// This is because pack.toml contains important information like modloader versions

/**
 * Parses a Packwiz index file and returns its contents.
 *
 * @param indexFilePath - The path of the TOML file.
 * @returns The parsed index as a structured object.
 */
export async function parsePackwizIndex(
  indexFilePath: string,
): Promise<PackwizIndex> {
  const indexFile = new Resource(indexFilePath);
  const parsed = toml.parse(await indexFile.fetchContents());
  const hashFormat = isValidHashFormat(parsed["hash-format"]);
  return {
    location: indexFile,
    hashFormat: hashFormat,
    files: parsed.files.map(
      (file: {
        file: string;
        hash: string;
        hashFormat: HashFormat;
        alias?: string;
        metafile?: boolean;
        preserve?: boolean;
      }) => {
        return new IndexFileEntry(
          indexFile.parent.join(file.file),
          file.hash,
          file.hashFormat || hashFormat,
          file.alias,
          file.metafile ?? false,
          file.preserve ?? false,
        );
      },
    ),
  };
}
